/*----------------------------------------------------------
 function  [X, S, y, Z, out] = mexAltSDP_BIQ(X, C,opts)
 
 Alternating direction augmented Lagrangian method 
  on computing theta problem

%   min  C'*X
%   s.t. X_ii - X_ni  = 0, i=1,...,n-1,  ==> scale by sqrt(2/3)
%        X_nn = 1,
%        X is pos. def. and nonnegative
%   
% Facts:
% 1) before scale: A*A' = diag(1.5, ..., 1.5, 1)
%     after scale: A*A' = I
%   

 input:
 * X:      initial solution
 * C:      objective coefficient
 * opts:    options
 * 

 output:
 * X:       primal solution
 * S:       dual variable
 * y:       Lagrangian multiplier for linear constraints
 * Z:       Lagrangian multiplier for nonnegativity constraints
 * out:    information 
  
 Auther: Zaiwen Wen
 Date:   08-01-2009 
 *
 *
 * !!!! be careful, interger and double variable
 * 2/3 is integer while 2./3. is double
 ----------------------------------------------------------*/

#include "AltSDP_util.h"
#include "mex.h"


/*  AX = diag(X); AX(1:n-1) = sqrt(2/3)*(AX(1:n-1) - X(1:n-1,n));  */
void ComputeAX(int n, double *X, double *y)
{
    int di, dl = (n-1)*n;
    for(di=0; di<n-1; di++)
    {
        y[di] = sqrt(2./3.)*( X[di*n+di] - X[dl+di]);
        
        /* mexPrintf("Xii: %e,  Xni: %e, yi: %e, %e,  %e\n",X[di*n+di] ,  X[dl+di], y[di],sqrt(2./3.)*( X[di*n+di] - X[dl+di]), ( X[di*n+di] - X[dl+di]) ); */
    }
    
    y[n-1] = X[n*n-1];
}


/* AX = diag(X); AX(1:n-1) = sqrt(2/3)*(AX(1:n-1) - X(1:n-1,n));  */
void ComputeASZ(int n, double *S, double *Z, double *y)
{
    int di, dl = (n-1)*n;
    for(di=0; di<n-1; di++)
        y[di] = sqrt(2./3.)*( S[di*n+di] - S[dl+di] + Z[di*n+di] - Z[dl+di]  );
    
    y[n-1] = S[n*n-1] + Z[n*n-1] ;
}


/*
ATy = spdiags(y,0,n,n);
ATy(n,1:n-1) = -0.5*y(1:n-1)';
ATy(1:n-1,n) = -0.5*y(1:n-1);
ATy = sqrt(2/3)*ATy;  ATy(n,n) = y(n);
 */
/* Here, we only store the upper triangle of the matrix ATy*/
void ComputeATy(int n, double *y, double *ATy)
{
    int di, dl = (n-1)*n ;
    double dtmp;
    for(di=0; di<n-1; di++)
    {
        dtmp = sqrt(2./3.)*y[di];
        ATy[di*n+di] = dtmp;
        ATy[dl+di] = -dtmp/2.;
    }
    ATy[n*n-1] = y[n-1];
}

void ComputeW(int n, double *C, double *ATy, double mu, double *X, double *Z, double *W, double *WW, bool compEigS)
{
    unsigned int di, dj, idx;
    if(compEigS)
    {
        for( di=0; di <n; di++)
        {
            for (dj = di; dj<n; dj++)
            {
                idx = dj*n+di;
                W[idx] = C[idx] -ATy[idx] -Z[idx] - mu*X[idx]; WW[idx] = W[idx];
            }
        }
    }
    else
    {
        for( di=0; di <n; di++)
        {
            for (dj = di; dj<n; dj++)
            {
                idx = dj*n+di;
                W[idx] = ATy[idx] +Z[idx]+ mu*X[idx] -C[idx];   WW[idx] = W[idx];
            }
        }
    }
    
}





/* The gateway routine */
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
  /* Input */
    double  *XX, *C;
    int     m, n;
    
    /* fields for options */
    double  ftol, mu, rmu, mu_min, mu_max, rho, delta_mu_l, delta_mu_u;
    int     record,  mxitr, min_mu_itr, max_mu_itr, max_itr_stag_lev1, max_itr_stag_lev2, max_itr_stag_lev3;
    
  /* Output */
    double *X, *S, *Z, *y;
    double *cycle, *flag;
    
    mxArray *fout[7];
    const char *fnames[] = {"pobj", "dobj", "gap", "pinf", "dinf", "itr",  "exit"};
    const char *fexit[] = {"optimal", "exceed max iteration"};
    int   nfields = 7;
    double *ppobj, *pdobj, *pgap, *ppinf, *pdinf, *pitr;
    
 /* work variable */
    int   itr, di, dj, idx, itmu_pinf , itmu_dinf, itr_stag_lev1=0, itr_stag_lev2=0, itr_stag_obj=0, itr_stag = 0;
    unsigned int nsq;
    double *tmp, b, *AC, *ATy, *W, *resi, nrmC, nrmb, objscale, pinf, dinf, gap, dtmp;
    double ref_inf, rel_pobj, rel_dobj, pobj_old, dobj_old;
    
    int    compEigS, TcompEigS;
    
    /*variables for eigenvalue decomposition */
    double *WW, *V, *D, *work;
    int nev, lwork, *ifail, *iwork;
    
    
    double alpha, beta;
    int inc=1, incy;
    char *NTRANS = "N", *TTRANS = "T", *uplo="U";;
    
    
  /*  Check for proper number of arguments. */
  /* NOTE: You do not need an else statement when using
     mexErrMsgTxt within an if statement. It will never
     get to the else statement if mexErrMsgTxt is executed.
     (mexErrMsgTxt breaks you out of the MEX-file.)
   */
    if (nrhs != 3)
        mexErrMsgTxt("Three inputs required.");
    if (nlhs != 5)
        mexErrMsgTxt("Five outputs required.");
    
  /* Create a pointer to the input matrix W, M, T. */
    if(mxIsSparse(prhs[0])) {mexErrMsgTxt("Matrix X must be a full matrix."); return;}
    if(mxIsSparse(prhs[1])) {mexErrMsgTxt("Matrix E must be a full matrix."); return;}
    
    
    XX = mxGetPr(prhs[0]);   C = mxGetPr(prhs[1]);
    m  = mxGetM(prhs[0]);    n = mxGetN(prhs[0]);
    if (m!=n) mexErrMsgTxt("input C should be a nxn matrix!");
    nsq = n*n;
    
    
    
    /* mexPrintf("n: %d, m: %d\n", n, m);  */
    
    /*Get fields from opts */
    if (mxGetField(prhs[2], 0, "record") == NULL )         record = 0;
    else    {   tmp = mxGetPr(mxGetField(prhs[2], 0, "record"));    record      = (int)*tmp;}

    if (mxGetField(prhs[2], 0, "mxitr") == NULL )         mxitr = 100;
    else    {   tmp = mxGetPr(mxGetField(prhs[2], 0, "mxitr"));       mxitr       = (int)*tmp;    }
   
    if (mxGetField(prhs[2], 0, "min_mu_itr") == NULL )        min_mu_itr = 0;    
    else    {        tmp = mxGetPr(mxGetField(prhs[2], 0, "min_mu_itr"));    min_mu_itr   = (int)*tmp;    }    
    
    if (mxGetField(prhs[2], 0, "max_mu_itr") == NULL )        max_mu_itr = 0;    
    else    {        tmp = mxGetPr(mxGetField(prhs[2], 0, "max_mu_itr"));    max_mu_itr   = (int)*tmp;    }
    
    if (mxGetField(prhs[2], 0, "mu") == NULL )        mu = 1.;
    else    {        tmp = mxGetPr(mxGetField(prhs[2], 0, "mu"));        mu          = *tmp;     }
        
    if (mxGetField(prhs[2], 0, "rmu") == NULL )        rmu = 0.1;  
    else    {        tmp = mxGetPr(mxGetField(prhs[2], 0, "rmu"));       rmu         = *tmp;    }
    
    if (mxGetField(prhs[2], 0, "mu_min") == NULL )        mu_min = 1;     
    else     {         tmp = mxGetPr(mxGetField(prhs[2], 0, "mu_min"));        mu_min         = *tmp;      }
    
    if (mxGetField(prhs[2], 0, "mu_max") == NULL )        mu_max = 100;
    else    {        tmp = mxGetPr(mxGetField(prhs[2], 0, "mu_max"));        mu_max         = *tmp;      }

    if (mxGetField(prhs[2], 0, "ftol") == NULL )        ftol = 1e-4;    
    else    {        tmp = mxGetPr(mxGetField(prhs[2], 0, "ftol"));      ftol        = *tmp;    }
    
    if(mxGetField(prhs[2], 0, "rho") == NULL )        rho = 1.6;
    else    {        tmp = mxGetPr(mxGetField(prhs[2], 0, "rho"));       rho     = *tmp;    }
    
    if (mxGetField(prhs[2], 0, "max_itr_stag_lev1") == NULL )        max_itr_stag_lev1 = 10;
   else    {        tmp = mxGetPr(mxGetField(prhs[2], 0, "max_itr_stag_lev1"));    max_itr_stag_lev1   = (int)*tmp;    }
    
    if (mxGetField(prhs[2], 0, "max_itr_stag_lev2") == NULL )        max_itr_stag_lev2 = 100;
    else    {        tmp = mxGetPr(mxGetField(prhs[2], 0, "max_itr_stag_lev2"));    max_itr_stag_lev2   = (int)*tmp;    }
    
    if (mxGetField(prhs[2], 0, "max_itr_stag_lev3") == NULL )        max_itr_stag_lev3 = 1000;
    else    {        tmp = mxGetPr(mxGetField(prhs[2], 0, "max_itr_stag_lev3"));    max_itr_stag_lev3   = (int)*tmp;    }  
    
    if (mxGetField(prhs[2], 0, "delta_mu_l") == NULL )         delta_mu_l = 1.;
    else    {        tmp = mxGetPr(mxGetField(prhs[2], 0, "delta_mu_l"));       delta_mu_l =  *tmp;    }
    
    if (mxGetField(prhs[2], 0, "delta_mu_u") == NULL )         delta_mu_u = 1.;
    else    {        tmp = mxGetPr(mxGetField(prhs[2], 0, "delta_mu_u"));       delta_mu_u = *tmp;    } 
    
    
     /*   
    mexPrintf("max_itr_stag_lev1: %d,  max_itr_stag_lev2: %d, max_itr_stag_lev3: %d\n", max_itr_stag_lev1, max_itr_stag_lev2, max_itr_stag_lev3);
      */

    if(record>=10)
        mexPrintf("mxitr: %d, ftol: %3.2e, mu: %3.2e, rmu: %3.2e, mu_min: %3.2e\n", mxitr, ftol, mu, rmu, mu_min);
    
  /* Set the output pointer to the output variable */
    plhs[0] = mxCreateDoubleMatrix(n, n, mxREAL);      X  = mxGetPr(plhs[0]);
    plhs[1] = mxCreateDoubleMatrix(n, n, mxREAL);      S  = mxGetPr(plhs[1]);
    plhs[2] = mxCreateDoubleMatrix(m, 1, mxREAL);      y  = mxGetPr(plhs[2]);
    plhs[3] = mxCreateDoubleMatrix(n, n, mxREAL);      Z  = mxGetPr(plhs[3]);
    
    dcopy(&nsq, XX, &inc, X, &inc);
    
    
    plhs[4] = mxCreateStructMatrix(1, 1, nfields, fnames);
    /* char *fnames[] = {"pobj", "dobj", "nrmr", "itr",  "exit"}; */
    fout[0] = mxCreateDoubleMatrix(1,1, mxREAL);   ppobj = mxGetPr(fout[0]);
    fout[1] = mxCreateDoubleMatrix(1,1, mxREAL);   pdobj = mxGetPr(fout[1]);
    fout[2] = mxCreateDoubleMatrix(1,1, mxREAL);   pgap = mxGetPr(fout[2]);
    fout[3] = mxCreateDoubleMatrix(1,1, mxREAL);   ppinf = mxGetPr(fout[3]);
    fout[4] = mxCreateDoubleMatrix(1,1, mxREAL);   pdinf = mxGetPr(fout[4]);
    fout[5] = mxCreateDoubleMatrix(1,1, mxREAL);   pitr  = mxGetPr(fout[5]);
    fout[6] = mxCreateString(fexit[0]);
    
    ATy     = mxCalloc(nsq, sizeof(double));
    W       = mxCalloc(nsq, sizeof(double));
    AC      = mxCalloc(m, sizeof(double));
    resi    = mxCalloc(m, sizeof(double));
    
    
    /* allocate variable for dsyevx */
    lwork   = 8*n;
    WW      =  mxCalloc(nsq, sizeof(double));
    V       =  mxCalloc(nsq,sizeof(double));
    D       =  mxCalloc(n, sizeof(double));
    work    =  mxCalloc(lwork,sizeof(double));
    iwork   =  mxCalloc(5*n, sizeof(int));
    ifail   =  mxCalloc(n, sizeof(int));
    
    
    /*
    ones(m,  y);
    ComputeATy(n, m, y, LE, ATy);
    printMat(n, n, ATy);
     
    ComputeAX(n, m, ATy, LE, y);
    printMat(m, 1, y);
     */
    
    /*scaling data*/
    /*
    b = 1;
    nrmC = dnrm2(&nsq, C, &inc);  dtmp = 1 + nrmC;  alpha = 1/dtmp;
    dscal(&nsq, &alpha, C, &inc);
    objscale = dtmp;  nrmC = 1+nrmC/dtmp;  nrmb = 1+b;
     */
    
    /* nrmC = 1 + matrix_norm(n, C, 1);    objscale = 1; */
    

    nrmC = matrix_norm(n, C, 3);  dtmp = 1 + nrmC;  alpha = 1/dtmp;
    dscal(&nsq, &alpha, C, &inc);    objscale = dtmp;  nrmC = 1+nrmC/dtmp;
    b = 1;   nrmb = 1+b;

    
    /*only compute the upper triangle*/
    zeros(m,  y);   zeros(nsq, S);  zeros(nsq, Z);   zeros(nsq, ATy);   zeros(n, AC);  zeros(n, resi);
    *ppobj =  objscale* ddotSYM(n, X, C);    *pdobj = 0.0;
    gap = fabs(*ppobj-*pdobj)/(1.+fabs(*ppobj)+fabs(*pdobj));
    ComputeAX(n, C, AC);
    
    /* resi = AX - b*/
    ComputeAX(n, X, resi);   resi[m-1]-=b;   pinf = dnrm2(&m, resi, &inc)/nrmb;  dinf = 1.;
    compEigS = true;
    
    itmu_pinf = 0;  itmu_dinf = 0;      itr_stag = 0;
    ref_inf = MAX(MAX(pinf, dinf), gap);
    
    
    
    if (record >= 1)
        mexPrintf("%4d   %3.2e   %+3.2e   %+3.2e   %+3.2e   %3.2e   %+3.2e  %4d (%d)\n",
        1, mu,  *ppobj,  *pdobj,  gap, pinf, dinf,  0, compEigS);
    
    for(itr = 2; itr<=mxitr; itr++)
    {
        pobj_old = *ppobj;  dobj_old = *pdobj;
        
         /* step 1: compute y
        y =  (AC - ComputeAX(n, S+Z, T) + mu*resi);  % compute dual y
          */
        
        ComputeASZ(n, S, Z, y);
        /*
        mexPrintf("AC:\n"); printVec(m, AC);
        mexPrintf("y:\n"); printVec(m, y);
        mexPrintf("resi:\n"); printVec(m, resi); */
        
        for(di=0; di<m; di++)
            y [di] = AC[di]-y[di]-mu*resi[di];
        /*
        alpha = -1;  daxpy(&m, &alpha, AC, &inc, y, &inc);  dscal(&m, &alpha, y, &inc);
        alpha = -mu; daxpy(&m, &alpha, resi, &inc, y, &inc);
         */
        
    /*      mexPrintf("y:\n"); printVec(m, y);  */
        
        
       /* step 2: compute Z */
        ComputeATy(n,  y, ATy);
        
        
        
        /*   W =  C - ATy -S- mu*X;  WPos = (W>0);       Z = zeros(n); Z(WPos) = W(WPos); */
       /* for(di=0; di<nsq; di++)
        {
            dtmp = c-ATy[di]-S[di]-mu*X[di];
            if(dtmp>0.) Z[di] = dtmp; else  Z[di] = 0.0;
        } */
        
        for( di=0; di <n; di++)
        {
            for (dj = di; dj<n; dj++)
            {
                idx = dj*n+di;   dtmp = C[idx]-ATy[idx]-S[idx]-mu*X[idx];
                if(dtmp>0.0)
                    Z[idx] = dtmp;
                else
                    Z[idx] = 0.0;
            }
        }
        
         /* step 3: compute S */
        
        /* compute S by projection */
        /* W =  C - ATy-Z - mu*X; */
        /*need an extra WW since it will be destoyed */
        ComputeW(n, C, ATy, mu, X, Z, W, WW, compEigS);

        TcompEigS = compEigS;
        AltSDPPROJ(n, nsq, mu, S, X, rho, XX, &compEigS, W,  WW, 1e-12, 1e15, 1e-8, V, D, &nev, lwork, work, iwork, ifail);


        
        
    /* -------------------------check optimality-----------------*/
       /*  *ppobj = objscale*ddot(&nsq,X, &inc, C, &inc); */
        *ppobj =  objscale* ddotSYM(n, X, C);
        *pdobj = objscale* b*y[m-1];
        gap = fabs(*ppobj-*pdobj)/(1+fabs(*ppobj)+fabs(*pdobj));
        ComputeAX(n, X, resi);   resi[m-1]-=b;   pinf = dnrm2(&m, resi, &inc)/nrmb;
        dinf = 0; /* norm(ATy+S+Z-C)/nrmC */
        for( di=0; di <n; di++)
        {
            idx = di*n+di;  dtmp = ATy[idx] + Z[idx] + S[idx] - C[idx]; dinf += dtmp*dtmp;
            for (dj = di+1; dj<n; dj++)
            {
                idx = dj*n+di;  dtmp = ATy[idx] + Z[idx]  + S[idx] - C[idx]; dinf += 2*(dtmp*dtmp);
            }
        }
        dinf = sqrt(dinf)/nrmC;
        
        dtmp = MAX(MAX(pinf,dinf), gap);
        /* dtmp = MAX(pinf,dinf); */
        
        if (record >= 1)
            mexPrintf("%4d   %3.2e   %+3.2e   %+3.2e   %+3.2e   %3.2e   %+3.2e  %4d (%d) (%3.2e,   %+3.2e)  %4d  %3.2e\n",
            itr, mu,  *ppobj,  *pdobj,  gap, pinf, dinf,  nev, TcompEigS, ref_inf, dtmp, itr_stag, pinf/dinf);
        
        /* optimal solution */
        if( dtmp <= ftol )
            break;
        
        /* detection of stagnation */
        
        if (dtmp < ref_inf)
        {
            ref_inf = dtmp;  itr_stag = 0;
        }
        else
        {
            itr_stag = itr_stag + 1;
        }
        
        if ( ( itr_stag > max_itr_stag_lev1 &&  dtmp<= ftol*10) 
            || ( itr_stag > max_itr_stag_lev2 &&  dtmp<= ftol*100) 
            || ( itr_stag > max_itr_stag_lev3 && dtmp <= ftol*1000) )
            break;
        
       
        /* update mu */
	/* note: the scaling was changed to 1-norm */
	/*
        if (pinf <= dinf)
        {
            itmu_pinf = itmu_pinf + 1;  itmu_dinf = 0; 
            if(itmu_pinf > max_mu_itr)
            {
                mu = MAX(mu*rmu, mu_min); itmu_pinf = 0;
            }
        }
        else
        {
            itmu_dinf = itmu_dinf + 1;   itmu_pinf = 0; 
            if(itmu_dinf > max_mu_itr)
            {
                mu = MIN(mu/rmu, mu_max); itmu_dinf = 0;
            }
        }
	*/	

        dtmp = pinf/dinf;
        if (dtmp <= delta_mu_l)
        {
            itmu_pinf = itmu_pinf + 1;  itmu_dinf = 0;
            if(itmu_pinf > max_mu_itr)
            {
                mu = MAX(mu*rmu, mu_min); itmu_pinf = 0;
            }
        }
	
        else if (dtmp> delta_mu_u)
        {
            itmu_dinf = itmu_dinf + 1;  itmu_pinf = 0;
            if(itmu_dinf > max_mu_itr)
            {
                mu = MIN(mu/rmu, mu_max); itmu_dinf = 0;
            }
        }     


        /*
        if (pinf < dinf )
            mu = MAX(mu*rmu, mu_min);
         */

        
    }
    
    
    *pgap= gap; *ppinf = pinf;  *pdinf = dinf;  *pitr = itr;
    /*mexPrintf("itr: %d, mxitr: %d\n", itr, mxitr); */
    
    if(itr>=mxitr)
    {
        fout[6] = mxCreateString(fexit[1]);     *pitr = itr-1;
    }
    
    for(di=0; di < nfields; di++)
        mxSetField(plhs[4], 0, fnames[di], fout[di]);
    
    
    /* scaling back solution */
    for(di=0; di<n-1; di++)
        y[di] = objscale*sqrt(2./3.)*y[di];
    y[n-1] = objscale*y[n-1];
   
    for( di=0; di <n; di++)
    {
        idx = di*n+di;  S[idx] = objscale*S[idx];   Z[idx] = objscale*Z[idx];
        for (dj = 0; dj<di; dj++)
        {
            idx = dj*n+di; itr = di*n+dj;     X[idx] = X[itr];  
            S[itr] = objscale*S[itr];  S[idx] = S[itr];      Z[itr] = objscale*Z[itr];  Z[idx] = Z[itr];
        }
    }         
    
    mxFree(ATy);   mxFree(W);   mxFree(AC);  mxFree(resi);
    mxFree(WW);    mxFree(V);   mxFree(D);    mxFree(work); mxFree(iwork); mxFree(ifail);
    
    
}
